﻿// Chart loading states enumeration
LoadStates = {
    NOT_STARTED: 0,
    FAST_LOADING: 1,
    AUTO_REFRESH: 2,
    COMPLETE: 3
};

// Data request rate in FAST_LOADING state
var FAST_RATE = 50;
// Data request rate in AUTO_REFRESH state
var NORMAL_RATE = 60000;
// Data request rate if error occurs in any loading state
var ERROR_RATE = 10000;
// Minimal time step, 1 ms
var TIME_EPSILON = 1 / 86400000;
// Site root path relative to ChartPro.aspx
var ROOT_PATH = "../../";

// Current loading state
var loadState = LoadStates.NOT_STARTED;
// Time offset of requested data
var timeOffset = 0;
// Last request is successful
var lastReqSuccessful = true;
// Input channel numbers retrieved from the query string
var cnlNumsParam = "";
// View IDs retrieved from the query string
var viewIDsParam = "";
// Date of the last requested data in string format
var reqDateStr = "";
// Current server date and time in string format
var nowStr = "";
// The chart
var chart = null;
// Displayed data point count
var dataPtCnt = 0;
// Popup dialogs manipulation object
var popup = new scada.Popup();

// The variables below must be defined in ChartPro.aspx
// Start date of the charting period in string format
var startDateStr = startDateStr || "";
// End date of the charting period in string format
var endDateStr = endDateStr || "";
// Current server date in string format
var todayStr = todayStr || "";
// Localized phrases
var phrases = phrases || {};

// Start chart loading process
function startLoadingChart() {
    if (startDateStr && todayStr) {
        cnlNumsParam = scada.utils.getQueryParam("cnlNums");
        viewIDsParam = scada.utils.getQueryParam("viewIDs");
        reqDateStr = startDateStr;
        continueLoadingChart();
    }
}

// Continue chart loading process
function continueLoadingChart() {
    if (lastReqSuccessful) {
        // determine loading state and time offset
        if (reqDateStr < endDateStr) {
            loadState = LoadStates.FAST_LOADING;
            timeOffset = Math.floor(timeOffset) + 1;
        } else {
            if (reqDateStr == todayStr) {
                loadState = LoadStates.AUTO_REFRESH;
                timeOffset = getNextTime();
            } else {
                loadState = LoadStates.COMPLETE;
            }

            if (nowStr) {
                showStatus(phrases.Generated + nowStr, true);
            }
        }

        logLoadingState();
    } else {
        showStatus(phrases.LoadError, false);
    }

    // enqueue chart loading
    switch (loadState) {
        case LoadStates.FAST_LOADING:
            setTimeout(loadChart, lastReqSuccessful ? FAST_RATE : ERROR_RATE);
            break;
        case LoadStates.AUTO_REFRESH:
            setTimeout(loadChart, lastReqSuccessful ? NORMAL_RATE : ERROR_RATE);
            break;
    }
}

// Get next time point to request
function getNextTime() {
    var len = chart.chartData.timePoints.length;
    return len ? chart.chartData.timePoints[len - 1] + TIME_EPSILON : 0;
}

// Log loading state and time offset to console
function logLoadingState() {
    var stateName = "unknown";
    switch (loadState) {
        case LoadStates.NOT_STARTED:
            stateName = "NOT_STARTED";
            break;
        case LoadStates.FAST_LOADING:
            stateName = "FAST_LOADING";
            break;
        case LoadStates.AUTO_REFRESH:
            stateName = "AUTO_REFRESH";
            break;
        case LoadStates.COMPLETE:
            stateName = "COMPLETE";
            break;
    }

    console.info(scada.utils.getCurTime() + " Chart loading state is " + stateName + ", time offset is " + timeOffset);
}

// Load chart data
function loadChart() {
    var operation = "ChartSvc.svc/GetDailyChartData";
    showStatus(phrases.DataLoading, true);

    $.ajax({
        url: operation +
            "?cnlNums=" + cnlNumsParam +
            "&viewIDs=" + viewIDsParam +
            "&startDate=" + startDateStr +
            "&timeOffset=" + timeOffset,
        method: "GET",
        dataType: "json",
        cache: false
    })
    .done(function (data, textStatus, jqXHR) {
        try {
            var parsedData = $.parseJSON(data.d);
            if (parsedData.Success) {
                scada.utils.logSuccessfulRequest(operation);
                if (obtainChartData(parsedData)) {
                    lastReqSuccessful = true;
                } else {
                    console.warn(scada.utils.getCurTime() + " Received data were not correctly processed");
                    lastReqSuccessful = false;
                }
            } else {
                scada.utils.logServiceError(operation, parsedData.ErrorMessage);
                lastReqSuccessful = false;
            }
        }
        catch (ex) {
            scada.utils.logProcessingError(operation, ex.message);
            lastReqSuccessful = false;
        }
    })
    .fail(function (jqXHR, textStatus, errorThrown) {
        scada.utils.logFailedRequest(operation, jqXHR);
        lastReqSuccessful = false;
    })
    .always(function () {
        continueLoadingChart();
    });
}

// Obtain received chart data
function obtainChartData(parsedData) {
    var timePoints = chart.chartData.timePoints;
    var trends = chart.chartData.trends;
    var newTimePoints = parsedData.TimePoints;
    var newTrends = parsedData.TrendDataArr;

    if (newTimePoints.length) {
        var existingLen = timePoints.length;

        if (newTrends.length == trends.length &&
            (!existingLen || timePoints[timePoints.length - 1] < newTimePoints[0])) {
            // append the received chart data to the existing
            Array.prototype.push.apply(timePoints, newTimePoints);

            for (var i = 0; i < trends.length; i++) {
                Array.prototype.push.apply(trends[i].trendPoints, newTrends[i].TrendPoints);
            }

            // draw the received part of the chart
            chart.resume(existingLen);

            // add rows to data table if it is visible
            if ($("#divData").hasClass("show")) {
                generateDataHtml();
            }
        } else {
            return false;
        }
    }

    reqDateStr = parsedData.ReqDateStr;
    todayStr = parsedData.TodayStr;
    nowStr = parsedData.NowStr;
    return true;
}

// Show status text
function showStatus(text, success) {
    if (text) {
        var chartStatus = $(".chart-status")
        chartStatus.children("span").text(text);

        if (success) {
            chartStatus.removeClass("error");
        } else {
            chartStatus.addClass("error");
        }
    }
}

// Apply additional css styles in case of using iOS
function styleIOS() {
    if (scada.utils.iOS()) {
        scada.utils.styleIOS($("#divData"));
        $(".chart-wrapper").css("display", "inline-block");
        $("#hlExport").parent().addClass("hidden"); // hide Export menu item
    }
}

// Update layout of the form
function updateLayout() {
    // general form layout
    var divChartContent = $("#divChartContent");
    var divChart = $("#divChart");
    var divChartSplitter = $("#divChartSplitter");
    var divData = $("#divData");
    var winH = $(window).height();

    divChartContent.outerHeight(winH);
    divChart.outerHeight(winH);
    divChartSplitter.outerHeight(winH);
    divData.outerHeight(winH, true);

    if (divData.hasClass("show")) {
        divChart.outerWidth($(window).width() - divData.outerWidth() - $("#divChartSplitter").outerWidth());
    } else {
        divChart.outerWidth($(window).width());
    }

    // chart layout
    updateChartLayout();
}

// Update layout of the chart
function updateChartLayout() {
    var CHART_PADDING = 20;
    var divChart = $("#divChart");

    divChart.css("overflow", "hidden"); // prevent redundant scrollbars
    $("#cnvChart")
        .outerWidth(divChart.innerWidth() - CHART_PADDING)
        .outerHeight(divChart.innerHeight() - $("#divHeader").outerHeight() - CHART_PADDING);
    divChart.css("overflow", "");
}

// Generate data HTML elements if needed
function generateDataHtml() {
    var TrendPointIndexes = scada.chart.TrendPointIndexes;
    var timePoints = chart.chartData.timePoints;
    var trends = chart.chartData.trends;
    var timePtCnt = timePoints.length;

    if (timePtCnt > 0 && trends.length > 0) {
        var dataTable = $("#tblData");
        var createTableRequired = dataTable.length == 0;

        if (createTableRequired) {
            var dataTableHtml = "<table id='tblData'></table>";
            dataTable = $(dataTableHtml);

            // add table header
            var hdrRowHtml = "<tr class='hdr'><td><span>" + phrases.TimeCol + "</span></td>";
            for (let trend of trends) {
                hdrRowHtml += "<td><span>[" + trend.cnlNum + "]</span></td>";
            }
            hdrRowHtml += "</tr>";
            dataTable.append(hdrRowHtml);

            $(".table-wrapper").append(dataTable);
            $("#lblNoData").addClass("hidden");
        }

        // add trends data
        if (dataPtCnt < timePtCnt) {
            var newRows = new Array(timePtCnt - dataPtCnt);
            var rowInd = 0;
            var showDates = chart.timeRange.endTime > 1;
            var prevDate = dataPtCnt > 0 ? Math.floor(timePoints[dataPtCnt - 1]) : null;

            for (var timePtInd = dataPtCnt; timePtInd < timePtCnt; timePtInd++) {
                var timePoint = timePoints[timePtInd];

                if (showDates) {
                    // add date row if date is changed
                    var curDate = Math.floor(timePoint);
                    if (prevDate != curDate) {
                        prevDate = curDate;
                        var dateRowHtml = "<tr class='date'><td>" + chart.dateToStr(timePoint) + "</td>";
                        for (let trend of trends) {
                            dateRowHtml += "<td></td>";
                        }
                        dateRowHtml += "</tr>";
                        newRows.length++;
                        newRows[rowInd] = dateRowHtml;
                        rowInd++;
                    }
                }

                // add data row
                var dataRowHtml = "<tr><td>" + chart.fastTimeToStr(timePoint, true) + "</td>";
                for (let trend of trends) {
                    var trendPoint = trend.trendPoints[timePtInd];
                    dataRowHtml += "<td style='color:" + trendPoint[TrendPointIndexes.COLOR_IND] + "'>" +
                        trendPoint[TrendPointIndexes.TEXT_IND] + "</td>";
                }
                dataRowHtml += "</tr>";
                newRows[rowInd] = dataRowHtml;
                rowInd++;
            }

            dataTable.append(newRows);
            dataPtCnt = timePtCnt;

            // create or update fixed table header
            if (createTableRequired) {
                setTimeout(scada.tableHeader.create.bind(scada.tableHeader), 0);
            } else {
                setTimeout(scada.tableHeader.update.bind(scada.tableHeader), 0);
            }
        }
    }
}

// Show modal dialog to select the chart channels
function showCnlsModal() {
    popup.showModal("ChartCnls.aspx" + window.location.search,
        new scada.ModalOptions([scada.ModalButtons.OK, scada.ModalButtons.CANCEL]),
        function (dialogResult, extraParams) {
            if (dialogResult) {
                // reload the chart with the selected channels
                var year = scada.utils.getQueryParam("year");
                var month = scada.utils.getQueryParam("month");
                var day = scada.utils.getQueryParam("day");
                var date = year && month && day ? new Date(year, month - 1, day) : null;
                var period = scada.utils.getQueryParam("period");
                window.location = ROOT_PATH +
                    scada.chart.dialog.getChartUrl(extraParams.cnlNums, extraParams.viewIDs, date, period);
            }
        });
}

// Show modal dialog to change the chart time period
function showPeriodModal() {
    popup.showModal("ChartPeriod.aspx" + window.location.search,
        new scada.ModalOptions([scada.ModalButtons.OK, scada.ModalButtons.CANCEL]),
        function (dialogResult, extraParams) {
            if (dialogResult) {
                // reload the chart with the selected start date and period
                var cnlNums = scada.utils.getQueryParam("cnlNums");
                var viewIDs = scada.utils.getQueryParam("viewIDs");
                window.location = ROOT_PATH +
                    scada.chart.dialog.getChartUrl(cnlNums, viewIDs, extraParams.date, extraParams.period);
            }
        });
}

// Show modal dialog contains export options
function showExportModal() {
    popup.showModal("ChartExport.aspx",
        new scada.ModalOptions([scada.ModalButtons.CLOSE], scada.ModalSizes.SMALL),
        function (dialogResult, extraParams) {
            if (dialogResult) {
                // export the chart
                switch (extraParams.exportFormat)
                {
                    case scada.ExportFormats.PDF:
                        scada.chart.export.exportToPdf();
                        break;
                    case scada.ExportFormats.PNG:
                        scada.chart.export.exportToPng();
                        break;
                    case scada.ExportFormats.EXCEL:
                        scada.chart.export.exportToExcel();
                        break;
                }
            }
        });
}

// Construct chart data markup and show it
function showData() {
    var MAX_DATA_WIDTH = 250;

    generateDataHtml();

    var divChartSplitter = $("#divChartSplitter");
    var divData = $("#divData");

    divChartSplitter.addClass("show");
    divData.addClass("show");

    if (!divData.data("alreadyShown")) {
        divData.data("alreadyShown", true);
        var dataWidth = divData.outerWidth() + scada.utils.getScrollbarWidth();
        divData.outerWidth(dataWidth <= MAX_DATA_WIDTH ? dataWidth : MAX_DATA_WIDTH);
    }

    updateLayout();
    chart.draw();
}

// Hide chart data
function hideData() {
    $("#divChartSplitter").removeClass("show");
    $("#divData").removeClass("show");
    updateLayout();
    chart.draw();
}

// Setup the Close button
function setupCloseButton() {
    if (scada.utils.isRealFullscreen()) {
        $("#btnClose").click(function () {
            window.close();
            return false;
        });
    } else {
        $("#btnClose").css("display", "none");
    }
}

// Bind events of the menu items
function bindMenuEvents() {
    $("#hlSelectCnls").click(function (event) {
        event.preventDefault();
        showCnlsModal();
    });

    $("#hlChangePeriod").click(function (event) {
        event.preventDefault();
        showPeriodModal();
    });

    $("#hlExport").click(function (event) {
        event.preventDefault();
        showExportModal();
    });

    $("#hlShowData").click(function (event) {
        event.preventDefault();
        $(this).addClass("hidden");
        $("#hlHideData").removeClass("hidden");

        setTimeout(showData, 0); // allow to close the dropdown before data are shown
    });

    $("#hlHideData").click(function (event) {
        event.preventDefault();
        $(this).addClass("hidden");
        $("#hlShowData").removeClass("hidden");

        hideData();
    });
}

$(document).ready(function () {
    // chart properties must be defined in ChartPro.aspx
    chart = new scada.chart.Chart($("#cnvChart"));
    chart.displaySettings = displaySettings;
    chart.timeRange = timeRange;
    chart.chartData = chartData;
    chart.phrases = phrases;

    scada.dialogs.rootPath = ROOT_PATH;
    styleIOS();

    setTimeout(function () {
        updateLayout();
        chart.draw();
        chart.bindHintEvents();
        chart.bindScalingEvents();
    }, 0); // timeout is needed to open chart in a popup window in Firefox
    
    scada.splitter.prepare(function () {
        updateChartLayout();
        chart.draw();
    });

    $(window).resize(function () {
        updateLayout();
        chart.draw();
    });

    setupCloseButton();
    bindMenuEvents();
    startLoadingChart();
});
